# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = [
    'TeamProcessorException',
    'NoAuthorizedTeamsException',
    'BaseProcessor',
    'UrlProcessor',
    'RegexpProcessor',
    'LaunchpadProcessor',
]

import datetime
import urllib
import urlparse
import re
import sys

from string import Template

from apache_openid.logging import debug, error


class TeamProcessorException(Exception):
    pass


class NoAuthorizedTeamsException(TeamProcessorException):
    pass


class BaseProcessor(object):
    '''Base Class for retrieving a list of authorized teams'''

    def __init__(self, request, options, session):
        self.request = request
        self.options = options
        self.session = session
        debug("Initializing processor: %s" % self.__class__.__name__)


    def get_authorized_teams(self):
        raise NotImplementedError

    def parse_authorized_teams(self, data):
        """Parse a string containing whitespace-separated team names."""
        teams = []
        for s in data.split():
            team = s.strip()
            if not team or team[0] == '#':
                continue
            teams.append(team)
        return teams


class OptionListProcessor(BaseProcessor):
    '''Parses a list of teams from the options list'''

    def get_authorized_teams(self):
        debug('Looking for authorized teams')
        if 'authorized-teams-list' in self.options:
            authorized_teams_raw = self.options.get('authorized-teams-list')
        elif 'authorized-teams' in self.options:
            # Allow the old option name for backwards compatibility
            authorized_teams_raw = self.options.get('authorized-teams')
        else:
            raise TeamProcessorException('No authorized-teams-list option')

        teams = self.parse_authorized_teams(authorized_teams_raw)
        debug('Found teams: %r', teams)
        return teams


class UrlProcessor(BaseProcessor):
    '''Retrieve a list of authorized teams from a URL'''

    def __init__(self, request, options, session):
        super(UrlProcessor, self).__init__(request, options, session)
        self.url = self.options.get('authorized-teams-list-url')

    def get_authorized_teams(self):
        debug('Looking for authorized teams')
        try:
            url_handle = urllib.urlopen(self.url)
            data = url_handle.read()
            url_handle.close()
            teams = self.parse_authorized_teams(data)
            debug('Found teams: %r', teams)
            return teams
        except IOError:
            error('Failed to fetch authorized list URL %r', self.url)
            raise TeamProcessorException(
                'Failed to fetch authorized list URL %r' % self.url)


class RegexpProcessor(BaseProcessor):
    '''Extracts a team name from a URL based on a regular expression match'''

    def __init__(self, request, options, session):
        super(RegexpProcessor, self).__init__(request, options, session)
        self.pattern = self.options.get('authorized-teams-regexp-pattern')
        self.target = self.options.get('authorized-teams-regexp-target', '$REQUEST_URI')
        self.group = self.options.get('authorized-teams-regexp-group', 0)
        try:
            self.group = int(self.group)
        except ValueError:
            pass

    @property
    def _substitutions(self):
        """A dictionary of substitutions.

        $-substitutions are allowed, using the CGI environment, plus the
        public attributes of the request object, plus session data, in that
        order of preference.
        """

        substitutions = self.request.subprocess_env.copy()
        for attribute in dir(self.request):
            if attribute.startswith('_'):
                continue
            value = getattr(self.request, attribute)
            if isinstance(value, basestring):
                substitutions[attribute] = value
        substitutions.update(self.session)
        debug('Substitutions: %s' % substitutions)
        return substitutions

    def get_authorized_teams(self):
        """Fetch a list of teams using regexp matching.

        This method performs a regular expression match (using Python's regexp
        syntax) against a string defined in the regexp-target option.  That
        option is a $-style string template which gets substitutions applied
        from the CGI environment, plus the public attributes of the request
        object, plus session data, in that order of preference.

        The regexp-group option names a regexp group that will return a team
        name when a successful match occurs.
        """
        debug('Looking for authorized teams')
        debug('Pattern: %r' % self.pattern)
        debug('target (raw): %r', self.target)
        target = Template(self.target).safe_substitute(self._substitutions)
        debug('target (cooked): %r', target)
        # Now do regular expression match to get the team name.
        try:
            match_object = re.search(self.pattern, target)
        except:
            error('Bad regular expression pattern: %r [target: %r]',
                       self.pattern, target)
            raise TeamProcessorException(
                'Bad regular expression pattern: %s' % self.pattern)

        if match_object is None:
            error('Pattern failed to match, ignoring.')
            raise NoAuthorizedTeamsException('Pattern failed to match')

        debug('Using group: %r', self.group)
        # Try to get the matched team by group name
        try:
            team = match_object.group(self.group)
        except IndexError:
            error('No group named %s' % self.group)
            raise NoAuthorizedTeamsException('No group named %s' % self.group)

        if not team:
            error('No team group value found')
            raise NoAuthorizedTeamsException('No team group value found')

        teams = [team]

        debug('Found teams: %r', teams)
        return teams


class LaunchpadProcessor(BaseProcessor):
    pass
